Set Implicit Arguments.
Require Import LibTactics.
Require Import LibLogic.
Require Import LibProd.
Require Import LibBag.
Require Import LibSet.
Require Import LibFunc.
Require Import LibRelation.
Local Notation path := rtclosure.
Require Import LibPer.
Require Import LibNatExtra.
Require Import LibStruct.
Require Import LibIter.
Require Import InverseNatNat.
Require Import Ackermann.
Require Import InverseAckermann.
Require Import MiscArith.
Require Import TLCBuffer.
Require Import UnionFind01Data.
Require Import UnionFind04Compress.
Require Import UnionFind05Rank.
Require Import UnionFind09Potential.
Require Import UnionFind10BasicEvolution.

(* -------------------------------------------------------------------------- *)

(* Assume [x] and [y] both have non-zero rank and both have a parent. Assume
   that there is a non-trivial path from [x] to [y]. Finally, assume that [x]
   and [y] have the same level, i.e., [k x = k y]. Let us say that [x] is
   pleasant if these conditions are met and unpleasant otherwise. *)

Definition pleasant V F K (x : V) :=
  0 < K x /\
  ~ is_root F x /\
  exists y,
  ~ is_root F y /\
  path F (p F x) y /\
  k F K x = k F K y.

(* -------------------------------------------------------------------------- *)

(* We define the displeasure of a vertex [x] as the number of unpleasant
   non-root ancestors of [x]. *)

Definition unpleasant_ancestors V F K (x : V) :=
  ancestors F x \n \set{ y | ~ pleasant F K y /\ ~ is_root F y }.

Definition displeasure V F K (x : V) :=
  card (unpleasant_ancestors F K x).

(* -------------------------------------------------------------------------- *)

(* A root has displeasure zero. *)

Lemma displeasure_of_root:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x,
  is_root F x ->
  displeasure F K x = 0.
Proof using.
  intros.
  unfold displeasure, unpleasant_ancestors.
  erewrite ancestors_of_root by eauto with is_dsf.
  rewrite disjoint_implies_empty_inter.
  eapply card_empty.
  rew_set in *. tauto.
Qed.

(* -------------------------------------------------------------------------- *)

(* The following lemmas relate the displeasure of a non-root vertex [x] and
   the displeasure of its parent [y]. *)

Lemma displeasure_parent_if_pleasant:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y,
  F x y ->
  pleasant F K x ->
  displeasure F K x = displeasure F K y.
Proof using.
  intros. unfold displeasure, unpleasant_ancestors. f_equal.
  erewrite ancestors_of_parent by eauto with is_dsf.
  eapply inter_union_disjoint_right.
  rew_set in *. tauto.
Qed.

Lemma displeasure_parent_if_unpleasant:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y,
  F x y ->
  ~ pleasant F K x ->
  displeasure F K x = displeasure F K y + 1.
Proof using.
  intros. unfold displeasure, unpleasant_ancestors.
  erewrite ancestors_of_parent by eauto with is_dsf.
  rewrite <- card_disjoint_union_single with (x := x).
  f_equal.
  eapply inter_union_subset_right.
  { rew_set in *. eauto using a_root_has_no_parent_contrapositive. }
  { eauto 6 with finite is_dsf. }
  { rew_set in *. rew_logic in *. eauto using ancestors_of_parent_disjoint with is_dsf. }
Qed.

Lemma displeasure_parent:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y,
  F x y ->
  displeasure F K x <= displeasure F K y + 1.
Proof using.
  intros.
  tests : (pleasant F K x).
  { erewrite displeasure_parent_if_pleasant by eauto. omega. }
  { erewrite displeasure_parent_if_unpleasant by eauto. omega. }
Qed.

(* -------------------------------------------------------------------------- *)

(* If [x] has non-zero rank, then the function [k] is well-defined on the set
   of the non-root ancestors of [x]. Let us write [levels] for the cardinal of
   the image of this set under [k]. That is, [levels] is the number of
   distinct levels of the non-root ancestors of [x]. *)

Definition levels V F K (x : V) :=
  card (image (k F K) (ancestors F x \n \set{ y | ~ is_root F y })).

(* -------------------------------------------------------------------------- *)

(* Because the image of the function [k] is included in the semi-open interval
   [[0, alpha n)], the number [levels F K x] is at most [alpha n]. *)

Lemma bounded_levels:
  forall V D F K x,
  @is_rdsf V D F K ->
  0 < K x ->
  levels F K x <= alpha (card D).
Proof using.
  intros. unfold levels.
  transitivity (card (interval_as_set 0 (alpha (card D)))).
  { eapply card_incl_le; [ eauto using finite_interval_as_set | ].
    unfold image, interval_as_set.
    eapply incl_prove. intros. do 2 (rew_set in *; unpack). subst. split; [ omega | ].
    forwards: ancestor_has_greater_rank; eauto.
    eauto using k_lt_alpha, at_most_one_root_ancestor with is_dsf omega.
  }
  rewrite card_interval_as_set; omega. (* DEPRECATED: using finite_interval_as_set *)
Qed.

(* -------------------------------------------------------------------------- *)

(* The following lemmas relate the levels of a non-root vertex [x] and the
   levels of its parent [y]. *)

Lemma levels_parent_if_pleasant:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y,
  F x y ->
  levels F K y <= levels F K x.
  (* one could prove an equality *)
Proof using.
  intros. unfold levels.
  eapply card_incl_le. { eauto 7 with finite is_dsf. }
  eapply image_covariant.
  eapply inter_covariant.
  eauto using ancestors_of_parent_inclusion with is_dsf.
  reflexivity.
Qed.

Lemma levels_parent_if_unpleasant:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y,
  F x y ->
  ~ pleasant F K x ->
  0 < K x ->
  levels F K y + 1 <= levels F K x.
  (* one could prove an equality *)
Proof using.
  intros. unfold levels.
  (* The key remark is that, because [x] is unpleasant, none of its non-root
     ancestors has the same level as [x]. *)
  assert (k F K x \notin image (k F K) (ancestors F y \n \set{ y | ~ is_root F y })).
  { unfold image, ancestors. intro. do 2 (rew_set in *; unpack).
    unfold pleasant in *. rew_logic in *.
    branches.
    { tauto. }
    { eauto using a_root_has_no_parent. }
    { match goal with h: forall x : V, _ |- _ => eapply h end.
      splits; eauto.
      erewrite parent_unique by eauto. assumption. }}
  (* The rest is just administration. *)
  erewrite <- card_disjoint_union_single by eauto 7 with finite is_dsf.
  eapply card_incl_le; [ eauto 7 with finite is_dsf | ].
  erewrite ancestors_of_parent with (x := x) by eauto with is_dsf.
  rewrite inter_union_subset_right by (rew_set in *; eauto using a_root_has_no_parent_contrapositive).
  rewrite image_union. rewrite image_singleton.
  reflexivity.
Qed.

(* -------------------------------------------------------------------------- *)

(* By definition of [pleasant], the number of unpleasant non-root ancestors
   of [x] is bounded by [levels F K x]. *)

Lemma bounded_displeasure_preliminary_1:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x z,
  path F x z ->
  is_root F z ->
  0 < K x ->
  displeasure F K x <= levels F K x.
Proof using.
  induction 2; intros.
  (* Base. *)
  { erewrite displeasure_of_root by eauto. omega. }
  (* Step. *)
  { tests : (pleasant F K x).
    { (* Case: [x] is pleasant. The displeasure of [x] is equal to the
         displeasure of its parent [y]. The set on the right-hand side
         is unchanged because, by definition of [pleasant], [x] admits
         a non-root ancestor with the same level. *)
      assert (hd: displeasure F K x <= displeasure F K y).
      { erewrite displeasure_parent_if_pleasant by eauto. eauto. }
      assert (hc: levels F K y <= levels F K x).
      { eauto using levels_parent_if_pleasant. }
      rewrite hd. rewrite <- hc. eauto using parent_has_nonzero_rank. }
    { (* Case: [x] is unpleasant. The displeasure of [x] is one more
         than the displeasure of its parent [y]. The measure [levels]
         decreases because, by definition of [pleasant], no non-root
         ancestor of [x] has the same level as [x]. *)
      assert (hd: displeasure F K x <= displeasure F K y + 1).
      { erewrite displeasure_parent_if_unpleasant by eauto. eauto. }
      assert (hc: levels F K y + 1 <= levels F K x).
      { eauto using levels_parent_if_unpleasant. }
      rewrite hd. rewrite <- hc.
      assert (forall a b, a <= b -> a + 1 <= b + 1). { intros. omega. }
      eauto using parent_has_nonzero_rank. }
  }
Qed.

(* -------------------------------------------------------------------------- *)

(* By combining the previous two results, we obtain the following. *)

Lemma bounded_displeasure_preliminary_2:
  forall V D F K x,
  @is_rdsf V D F K ->
  0 < K x ->
  displeasure F K x <= alpha (card D).
Proof using.
  intros.
  forwards [ z [ ? ? ]]: is_dsf_defined_is_repr x. eauto with is_dsf.
  rewrite bounded_displeasure_preliminary_1 by eauto.
  rewrite bounded_levels by eauto with is_repr.
  eauto.
Qed.

(* -------------------------------------------------------------------------- *)

(* Finally, by accounting for vertices of rank 0, we obtain the following. *)

Lemma bounded_displeasure:
  forall V D F K x,
  @is_rdsf V D F K ->
  displeasure F K x <= alpha (card D) + 1.
Proof using.
  intros.
  tests : (0 < K x).
  { rewrite bounded_displeasure_preliminary_2 by eauto. omega. }
  { tests : (is_root F x).
    { erewrite displeasure_of_root by eauto. omega. }
    { unfold is_root in *. rewrite not_forall_not in *. unpack.
      rewrite displeasure_parent by eauto.
      rewrite bounded_displeasure_preliminary_2 by eauto using parent_has_nonzero_rank.
      omega. }}
Qed.

(* -------------------------------------------------------------------------- *)

(* A step of path compression at [x] does not increase the displeasure of [y],
   the parent of [x]. This sounds obvious, but proving it is pretty dreary.
   This is the object of the following two lemmas. *)

Lemma compress_preserves_pleasant_above_y:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y z,
  F x y ->
  path F y z ->
  forall v,
  path F y v ->
  pleasant F K v ->
  pleasant (compress F x z) K v.
Proof using.
  unfold pleasant. intros. unpack.
  (* Name [av] the ancestor of [v] which has the same level. *)
  match goal with h: path _ _ ?toto |- _ => rename toto into av end.
  assert (x <> v).
  { eapply paths_have_distinct_endpoints; eauto with is_dsf tclosure. }
  assert (path F y (p F v)).
  { eapply rtclosure_trans_explicit; [ eauto | eapply rtclosure_once ].
    eapply parent_spec. eauto. }
  splits; eauto using compress_preserves_roots_converse.
  exists av. splits.
  { eauto using compress_preserves_roots_converse. }
  { eapply compress_preserves_paths_out_of_y; eauto with is_dsf.
    erewrite compress_does_not_change_parent_of_v; eauto.
    erewrite compress_does_not_change_parent_of_v; eauto. }
  { erewrite compress_preserves_k_above_y by eauto.
    erewrite compress_preserves_k_above_y by eauto using rtclosure_trans_explicit.
    assumption. }
Qed.

Lemma compress_preserves_displeasure_of_y:
  forall V D F K,
  @is_rdsf V D F K ->
  forall x y z,
  F x y ->
  path F y z ->
  displeasure (compress F x z) K y <= displeasure F K y.
Proof using.
  unfold displeasure, unpleasant_ancestors. intros.
  eapply card_incl_le. eauto 6 with is_dsf finite.
  (* We cannot apply [inter_covariant] and reason separately about each
     side of the intersection, as we need to keep track of the fact that
     we are above [y]. *)
  eapply incl_prove. unfold ancestors. intros. rew_set in *. unpack. splits;
  eauto 6 using compress_preserves_pleasant_above_y,
    compress_preserves_paths_converse, compress_preserves_roots with is_dsf.
Qed.

