Set Implicit Arguments.
Require Import LibTactics.
Require Import CFHeaps CFLibCredits. (* CFHeaps might not be needed *)
Require Import UnionFind_ml.
Require Import LibFunc LibSet LibRelation LibPer.
Require Import LibNatExtra.
Require Import InverseAckermann.
Require Import TLCBuffer.
Require Import UnionFind01Data.
Require Import UnionFind02EmptyCreate.
Require Import UnionFind03Link.
Require Import UnionFind04Compress.
Require Import UnionFind04IteratedCompression.
Require Import UnionFind05Rank.
Require Import UnionFind06RankEmptyCreate.
Require Import UnionFind07RankLink.
Require Import UnionFind08RankCompress.
Require Import UnionFind09Potential.
Require Import UnionFind12Evolution.


(* -------------------------------------------------------------------------- *)
(* Default credits to nat instead of Z *)

Notation "'\$' x" := (\$_nat x) 
  (at level 40, format "\$ x") : heap_scope.

Ltac xpay_core tt ::= xpay_nat_core tt.


(* -------------------------------------------------------------------------- *)

Section CardParameter.
  (* This section is used to parameterize everything over [N]. *)
  (* Also, it implies that all [Hints] are local, which is good. *)

(** The maximal number of vertices, used in the definition of the potential *)

Variable N : nat. 

(* Hints for the type-checker. *)

(* It seems important that [D] and [R], which are the public parameters to
   [UF], mention [elem] in their types, as opposed to [loc]. Ultimately, the
   type [elem] is made opaque. *)

Implicit Types D : set elem.         (* domain *)
Implicit Types F : binary elem.      (* edges *)
Implicit Types R : elem -> elem.     (* functional view of [is_repr F] *)
Implicit Types K : elem -> nat.      (* rank of every vertex *)
Implicit Types M : map elem content. (* CFML's view of the memory *)

(* Inhabited type instance, which ideally would be generated by CFML *)

Global Instance Inhab_content : Inhab content. 
Proof. intros. applys (prove_Inhab (Root 0)). Qed.


(* -------------------------------------------------------------------------- *)
(** * Representation predicate *)

(** Relation between the graph [D/F/K] and the map [M] that describes
    the memory. In short,
    1. the domain of [M] coincides with [D]; 
    2. the links in [M] encode [F]; 
    3. the ranks stored at every root in [M] agree with [K]. *)

Record Mem D F K M : Prop := {
  Mem_dom : dom M = D;
  Mem_val : forall x, x \in D -> 
    match M[x] with
    | Link y => F x y 
    | Root r => is_root F x /\ r = K x
    end
 }.

(** Relation between the exposed interface [D/R] and the graph [D/F/K]. *)

Record Inv D F K R : Prop := {
  Inv_rdsf : is_rdsf D F K;
  Inv_incl : incl_fr R (is_repr F); (* [R] agrees with [F] *)
  Inv_card : (card D <= N)%nat
  }.

(* [UF D R] is the representation predicate that asserts that the 
   instance of Union-Find has domain [D], and that [R] is a function
   that maps every vertex in [D] to its representative. *)

Definition UF D R : hprop :=
  Hexists F K M,
  Group (Ref Id) M \*
  \[ Inv D F K R ] \* \[ Mem D F K M ] \*
  \$ (Phi D F K N).


(* -------------------------------------------------------------------------- *)
(* Remark *)

(* A proof that [incl_fr R (is_repr F)] is a definition equivalent to
   [rel_le (rel_func R) (is_repr F)], which is shown in the paper. *)

Definition rel_func R := fun x y => y = R x.

Remark rel_func_equiv : forall R F,
  rel_le (rel_func R) (is_repr F) <-> incl_fr R (is_repr F).
Proof. 
  intros R F. generalize (is_repr F) as R'. intros R'. 
  unfold rel_func, incl_fr. iff H; intros x; specializes H x.
    applys* H.
    intros y Hy. subst~.
Qed.


(* -------------------------------------------------------------------------- *)
(* Hints *)

(* Use group specification for ml_get and ml_set *)

Hint Extern 1 (RegisterSpec ml_get) => Provide (@ml_get_spec_group content).
Hint Extern 1 (RegisterSpec ml_set) => Provide (@ml_set_spec_group content).

Hint Resolve Inv_rdsf Inv_incl Inv_card.
Hint Resolve is_rdsf_is_dsf.

Ltac hsimpl_hook H ::= (* needed for [hsimpl] to work well on [UF] *)
  match H with
  | UF _ _ => hsimpl_try_same tt 
  end.

Ltac math_0 ::= unfolds rank.
Hint Extern 1 (_ = _ :> rank) => math.
Hint Extern 1 (?x \in ?D) => congruence.
Hint Extern 1 (Inhab content) => apply Inhab_content.


(* -------------------------------------------------------------------------- *)
(** * Lemmas about [Mem] *)

Lemma Mem_root : forall D R F K M x,
  Inv D F K R -> Mem D F K M -> x \in D -> R x = x ->
  M[x] = Root (K x).
Proof.
  introv HI HM Dx Rx.
  forwards~ HV: (Mem_val HM) x. destruct (M[x]).
    false* (>> R_self_is_root HI HV).
    destruct HV. congruence. 
Qed.

Lemma Mem_root_inv : forall D F K M x rx,
  Mem D F K M -> x \in D -> M[x] = Root rx -> 
  is_root F x /\ rx = K x.
Proof using.
  introv HM Dx HE. forwards~ HVx: (Mem_val HM) x.
  destruct (M[x]); tryfalse. inverts* HE.
Qed.

Lemma Mem_make : forall D F K D' r M,
  Mem D F K M ->
  is_rdsf D F K ->
  r \notin D ->
  D' = D \u \{r} ->
  Mem D' F K (M[r:=Root 0]).
Proof using.
  introv [HD HV] HG Dr ED. constructors.
    subst D'. simpl. rewrite~ dom_update; congruence.
    subst D'. simpl. introv Dx. 
      rewrite update_read_if. case_if.
        split. 
          { eauto using only_roots_outside_D with is_dsf. }
          { (* We use the fact that [K] is zero outside the domain. *)
            erewrite is_rdsf_zero_rank_outside_domain by eauto. math. }
        set_in; tryfalse. forwards~ HM: HV x.
Qed.

Lemma Mem_link : forall D F K F' M x r,
  Mem D F K M ->
  x \in D ->
  F' = link F x r ->
  Mem D F' K (M[x:=Link r]).
Proof using.
  introv [HD HR] Dx EG'. subst. constructors; simpl.
  rewrite~ dom_update_index.
  intros a Da. specializes HR Da.
  rewrite update_read_if. case_if.
    right. split~. (* later: use a lemma *)
    destruct (M[a]).
      left~.
      destruct HR. split*. applys* is_root_link.
Qed.

Lemma Mem_link' : forall D F K F' M x r,
  Mem D F K M ->
  x \in D ->
  F' = compress F x r ->
  Mem D F' K (M[x:=Link r]).
Proof using.
  introv [HD HR] Dx EG'. subst. constructors; simpl.
  rewrite~ dom_update_index.
  intros a Da. specializes HR Da.
  rewrite update_read_if. case_if.
    right~.
    destruct (M[a]).
      left~.
      destruct HR. split*. applys* compress_preserves_roots_other_than_x.
Qed.

Lemma Mem_link_incr : forall D F K F' K' M x r (rx:rank),
  Mem D F K M ->
  x \in D ->
  r \in D -> 
  x <> r ->
  is_root F r ->
  rx = K x ->
  F' = link F x r ->
  K' = fupdate K r (1 + K x)%nat ->
  Mem D F' K' (M[x:=Link r][r:=Root (rx + 1)]).
Proof using.
  introv HM Dx Dr HN Hr Rx HK EG'.
  forwards* HM': Mem_link x r HM.
  sets M': (M[x:=Link r]). clearbody M'.
  destruct HM' as [HD HV]; simpls. constructors; simpl.
    rewrite~ dom_update_index. rewrite~ LibMap.index_def.
    intros a Da. specializes HV Da. rewrite update_read_if. case_if.
      split. 
        subst F'. applys* is_root_link.
        subst K'. subst rx. unfold fupdate. case_if~.
      destruct (M'[a]).
        subst F'. auto.
        subst F' K'. destruct HV. split~. unfolds fupdate. case_if~.
Qed.


(* -------------------------------------------------------------------------- *)
(** * Lemmas about the [Inv] and [Phi] *)

Lemma Inv_make : forall D R F K D' r,
  Inv D F K R ->
  (card D < N)%nat ->
  r \notin D ->
  D' = D \u \{r} ->
  Inv D' F K R.
Proof.
  introv [ Hrdsf Hroots Hcard ] HC Dr ED. subst. constructor.
  { eauto using is_rdsf_create. }
  { assumption. }
  { erewrite card_union_le. erewrite card_single. omega.
    unfolds* is_rdsf. applys finite_single. }
Qed.

Definition Inv_link_hyp D R F K x y :=
  Inv D F K R /\
  x <> y /\
  x \in D /\
  y \in D /\
  R x = x /\ 
  R y = y.

Definition Inv_link_conc D R F K x y z F' K' :=
  let R' := link_by_rank_R R x y z in
  Inv D F' K' R' /\ 
  exists (n : nat), (Phi D F K N + alpha N = Phi D F' K' N + n)%nat.

Lemma Inv_link: forall D R F K x y,
  Inv_link_hyp D R F K x y ->
  Inv_link_conc D R F K x y
    (* z  := *) (new_repr_by_rank K x y)
    (* F' := *) (link_by_rank_F F K x y)
    (* K' := *) (link_by_rank_K K x y).
Proof using.
  unfold Inv_link_hyp, Inv_link_conc.
  intros. unpack. split. constructor.
  (* Preservation of [is_rdsf]. *)
  { eauto 8 using is_rdsf_link, R_self_is_root. }
  (* Preservation of the agreement between [R] and [F]. *)
  { eauto 8 using link_by_rank_R_link_by_rank_F_agree, R_self_is_root. }
  (* Preservation of [card D <= N]. (Trivial.) *)
  { eauto. }
  (* Change in potential. *)
  { eapply leq_to_eq_plus.
    eapply potential_increase_during_link; eauto using R_self_is_root. }
Qed.

(* The lemmas below are too involved to be inlined. *)

Lemma Inv_link_lt : forall D R F K x y,
  Inv_link_hyp D R F K x y ->
  (K x < K y)%nat ->
  exists F', F' = link F x y /\
  Inv_link_conc D R F K x y y F' K.
Proof.
  intros. forwards* HL: Inv_link.
  unfold link_by_rank_F, new_repr_by_rank, link_by_rank_K in HL.
  repeat case_if; try omega. eauto.
Qed.

Lemma Inv_link_gt : forall D R F K x y,
  Inv_link_hyp D R F K x y ->
  (K x > K y)%nat ->
  exists F', F' = link F y x /\ 
  Inv_link_conc D R F K x y x F' K.
Proof.
  intros. forwards* HL: Inv_link.
  unfold link_by_rank_F, new_repr_by_rank, link_by_rank_K in HL.
  repeat case_if; try omega. eauto.
Qed.

Lemma Inv_link_eq : forall D R F K x y,
  Inv_link_hyp D R F K x y ->
  (K x = K y)%nat -> 
  exists F' K', F' = link F y x /\ K' = fupdate K x (1+K x)%nat /\
  Inv_link_conc D R F K x y x F' K'.
Proof.
  intros. forwards* HL: Inv_link.
  unfold link_by_rank_F, new_repr_by_rank, link_by_rank_K in HL.
  repeat case_if; try omega. eauto.
Qed.


(* -------------------------------------------------------------------------- *)
(** * Private lemmas on the representation predicate *)

(* If [UF D R] holds, then [Inv D F K R] holds for some [F] and [K]. *)

Lemma UF_Inv : forall D R,
  UF D R ==> UF D R \* Hexists F K, \[Inv D F K R].
Proof using.
  intros. unfold UF. hextract as F K M HI HM. hsimpl*.
Qed.

Lemma UF_extract : forall D R (P:Prop),
  (forall F K, Inv D F K R -> P) -> 
  UF D R ==> UF D R \* \[ P ].
Proof using.
  intros. unfold UF. hextract as F K M HI HM. hsimpl*.
Qed.


(* -------------------------------------------------------------------------- *)
(** * Public lemmas on the representation predicate *)

(* If [UF D R] holds, then [idempotent R] holds. *)

Lemma UF_idempotent : forall D R,
  UF D R ==> UF D R \* \[idempotent R].
Proof using.
  intros. applys UF_extract. eauto using idempotent_R.
Qed.

(* If [UF D R] holds, then [R] has its image in [D]. *)

Lemma UF_image : forall D R,
  UF D R ==> UF D R \* \[forall x, x \in D -> R x \in D].
Proof using.
  intros. applys UF_extract. eauto using sticky_R.
Qed.

(* If [UF D R] holds, then [R] is the identity outside of [D]. *)

Lemma UF_identity : forall D R,
  UF D R ==> UF D R \* \[forall x, x \notin D -> R x = x].
Proof using.
  intros. applys UF_extract. eauto using R_is_identity_outside_D.
Qed.

(** Initialization *)

Lemma UF_create :
  \[] ==> UF \{} id.
Proof using.
  unfold UF.
  hchange (>> Group_create (Ref (@Id content))).
  hsimpl (\{} : set loc) (@empty loc) (fun (_:loc) => arbitrary).
  simpl. rewrite Phi_empty. rewrite~ credits_nat_zero_eq_prove. 
  constructors.
    rewrite~ dom_empty.
    introv Dx. simpls. false. 
  constructors; simpl.
    apply is_rdsf_empty.
    { intro x. eapply is_repr_empty. }
    { rewrite card_empty. omega. }
Qed.


(* -------------------------------------------------------------------------- *)
(* Demo. *)

Theorem link_spec' : forall D R x y, 
  x \in D -> y \in D -> R x = x -> R y = y ->
  App UnionFind_ml.link x y; 
    (UF D R \* \$(alpha N+1)) 
    (fun z => UF D (link_by_rank_R R x y z)
              \* \[z = x \/ z = y]). 
Proof.
  introv Dx Dy Rx Ry. xcf_app. credits_split. xpay. xapps. xif.
  { xret. rewrite~ link_by_rank_R_same. hsimpl~. }
  { unfold UF at 1. xextract as F K M HI HM. lets HMD: (Mem_dom HM). 
    unfold elem in *. xapps~. xapps~. xmatch.
    { forwards* (HRx&Kx): Mem_root_inv x HM.
      forwards* (HRy&Ky): Mem_root_inv y HM. xif.
      { forwards* (F'&EF'&HI'&n&EQ): (>> Inv_link_lt D R F K x y). splits*. math.
        xchange (>> credits_nat_eq EQ). chsimpl. xapp*. xret. 
        unfold UF. chsimpl*. applys* Mem_link. }
      { xif. 
        { forwards* (F'&EF'&HI'&n&EQ): (>> Inv_link_gt D R F K x y). splits*. math.
          xchange (>> credits_nat_eq EQ). chsimpl. xapp*. xret. 
          unfold UF. chsimpl*. applys* Mem_link. }
        { asserts: (rx = ry). math. asserts: (K x = K y). math.
          forwards* (F'&K'&EF'&EK'&HI'&n&EQ): (>>Inv_link_eq D R F K x y). splits*.
          xchange (>> credits_nat_eq EQ). chsimpl*. xapp~. xapp~. 
            apply~ LibMap.indom_update_already.
          xret. rewrite H2 in EK'. unfold UF. chsimpl*. applys* Mem_link_incr HM. }}} 
    { xfail ((rm C0) (K x) (K y)). fequals; applys* Mem_root. } }
Qed.


(* -------------------------------------------------------------------------- *)
(* Verification. *)

(* Verification of [make]. *)

Lemma make_spec :
  Spec UnionFind_ml.make () |J>> forall D R, (card D < N)%nat ->
    J (UF D R \* \$ 1)
      (fun r => UF (D \u \{r}) R \* \[r \notin D] \* \[R r = r]).
Proof using.
  xcf. introv HD. xpay.
  unfold UF. xextract as F K M HI HM. xapp. intros r.
  hchange (Group_add_fresh r). typeclass. typeclass. hextract as Mr.
  asserts Dr: (r \notin D). rewrite~ <- (Mem_dom HM).
  forwards* EQ: Phi_extend N. rewrite EQ. hsimpl~.
  applys* R_is_identity_outside_D. applys* Mem_make. applys* Inv_make.
Qed.

(* Verification of [find]. *)

Lemma find_spec : 
  Spec UnionFind_ml.find x |J>> forall D R, x \in D -> 
    J (UF D R \* \$ (alpha N + 2))
      (fun (r:loc) => UF D R \* \[R x = r]).
Proof using. 
  asserts S': (Spec find x |J>> forall d D R F K F' M, 
    Inv D F K R -> Mem D F K M -> x \in D -> bw_ipc F x d F' -> 
    J (Group (Ref Id) M \* \$ (d+1))
      (fun r' => Hexists M', Group (Ref Id) M' \* 
         \[Mem D F' K M' /\ r' = R x])).
  { xinduction_heap Wf_nat.lt_wf. xcf.
    intros x d IH. hide IH. introv HI HM Dx HC. credits_split. xpay. 
    lets HMD: (Mem_dom HM).
    unfold elem in *.
    xapps*. forwards* HV: (Mem_val HM) x. 
    unfold elem in *.
    xmatch; rename H0 into HK; rewrite <- HK in HV.
    (* root *)
    { destruct HV as [HR HP]. inverts HC.
      forwards*: is_root_R_self HR. xret*.
      false* a_root_has_no_parent HR. }
    (* link *)
    { rename HV into HF. invert HC.
        intros. subst F. subst. false* a_root_has_no_parent.
      intros F'0 F'' x0 y0 z' d' HF' HR' HB' EF' Ex Ed EF''. subst F'' x0 d.
      lets* E: is_dsf_functional D HF HF'. subst y0. clear HF'.
      assert (y \in D). { eauto with confined is_dsf. }
      assert (z' = R y). { eapply is_repr_incl_R; eauto. } subst z'.
      xapp_spec~ IH (>> HI HB').
      replace (S d') with ((d'+1)%nat); [|omega]. chsimpl.
      intros M' (HM'&Er). lets HMD': (Mem_dom HM').
      unfold elem in *.
      xapp~. xret. hsimpl. split~. subst z.
      applys~ Mem_link' HM'.
      assert (is_equiv F x y). { eauto using path_is_equiv with rtclosure. }
      assert (R x = R y). { eauto using is_equiv_incl_same_R. } congruence.
    }
  }
  (* Main specification *)
  xweaken S'. clear S'. simpl. intros x. intros RS LRS KRS.
  introv Dx. unfold UF. xextract as F K M HI HM.
  forwards* (d&F'&HC&HP): amortized_cost_of_iterated_path_compression x N.
  change (2%nat) with (1+1)%nat.
  forwards (H'&E): credits_nat_le_rest HP.
  xchange E. chsimpl. credits_split. xgc H'. xframe (\$ Phi D F' K N).
  xapply (>> KRS HI HM Dx HC). change (S d) with (1+d)%nat. chsimpl.
  xok. clears RS. intros z. xextract as M' (HM'&Hz). subst z.
  intros. hsimpl~. 
  constructor; eauto using is_rdsf_bw_ipc, bw_ipc_preserves_RF_agreement.
Qed.

Hint Extern 1 (RegisterSpec UnionFind_ml.find) => Provide find_spec.

(* Verification of [equiv]. *)

Lemma equiv_spec : 
  Spec UnionFind_ml.equiv x y |J>> forall D R,
    x \in D -> y \in D ->
    J (UF D R \* \$(2*(alpha N) + 5)) 
      (fun b => UF D R \* \[b = isTrue (R x = R y)]). 
Proof using.
  xcf. introv Dx Dy. 
  math_rewrite (2*(alpha N)+5 = 1 + (alpha N+2) + (alpha N+2))%nat.
  credits_split. xpay.
  xapp* as x'. chsimpl. intros Rx.
  xapp* as y'. chsimpl. intros Ry.
  xapp. intros b. hsimpl. subst. rew_logic*. 
Qed.

(* Verification of [link]. *)

Lemma link_spec :
  Spec UnionFind_ml.link x y |J>> forall D R,
    x \in D -> y \in D -> R x = x -> R y = y ->
    J (UF D R \* \$(alpha N + 1)) 
      (fun z => UF D (link_by_rank_R R x y z)
                 \* \[z = x \/ z = y]).
Proof using.
  xcf. introv Dx Dy Rx Ry. credits_split. xpay. xapps. xif.
  (* case x == y *)
  xret. rewrite~ link_by_rank_R_same. hsimpl~. 
  (* case x != y *)
  unfold UF at 1. xextract as F K M HI HM. lets HMD: (Mem_dom HM).
  unfold elem in *.
  xapps~. xapps~. xmatch.
  (* case: x and y are roots *)
  forwards* (HRx&Kx): Mem_root_inv x HM.
  forwards* (HRy&Ky): Mem_root_inv y HM. xif.
  (* case rx < ry *)
  forwards* (F'&EF'&HI'&n&EQ): (>> Inv_link_lt D R F K x y). splits*. math.
  xchange (>> credits_nat_eq EQ). chsimpl. xapp*. xret. unfold UF. chsimpl*.
  applys* Mem_link.
  (* case rx > ry *)
  xif.
  forwards* (F'&EF'&HI'&n&EQ): (>> Inv_link_gt D R F K x y). splits*. math.
  xchange (>> credits_nat_eq EQ). chsimpl. xapp*. xret. unfold UF. chsimpl*.
  applys* Mem_link.
  (* case rx = ry *)
  asserts: (rx = ry). math. asserts: (K x = K y). math.
  forwards* (F'&K'&EF'&EK'&HI'&n&EQ): (>> Inv_link_eq D R F K x y). splits*.
  xchange (>> credits_nat_eq EQ). chsimpl. xapp~. xapp~. 
   apply~ LibMap.indom_update_already.
  xret. rewrite H2 in EK'. unfold UF. chsimpl*. applys* Mem_link_incr HM.
  (* case: x or y not a root *)
  xfail ((rm C0) (K x) (K y)). fequals; applys* Mem_root.
Qed.

Hint Extern 1 (RegisterSpec UnionFind_ml.link) => Provide link_spec.

(* Verification of [union]. *)

Lemma union_spec :
  Spec UnionFind_ml.union x y |J>> forall D R,
    x \in D -> y \in D ->
    J (UF D R \* \$(3*(alpha N) + 6)) 
      (fun z => UF D (link_by_rank_R R x y z) 
         \* \[z = R x \/ z = R y]). 
Proof using.
  xcf. introv Dx Dy. 
  math_rewrite (3*(alpha N)+6 = 1+(alpha N+1)+(alpha N+2)+(alpha N+2))%nat.
  credits_split. xpay.
  xapp* as x'. chsimpl. intros Rx.
  xapp* as y'. chsimpl. intros Ry.
  subst.
  xchange UF_Inv. xextract as F K HI. lets: sticky_R.
  forwards* HP: idempotent_R. hnf in HP.
  xapp*. chsimpl. intros r. hextract as E. 
  rewrites* (>> link_by_rank_root). hsimpl*.
Qed.

End CardParameter.


(* Make [elem] and [UF] abstract, as best as we can. *)

Global Opaque elem UF.

Hint Extern 1 (RegisterSpec UnionFind_ml.make) => Provide make_spec.
Hint Extern 1 (RegisterSpec UnionFind_ml.find) => Provide find_spec.
Hint Extern 1 (RegisterSpec UnionFind_ml.equiv) => Provide equiv_spec.
Hint Extern 1 (RegisterSpec UnionFind_ml.link) => Provide link_spec.
Hint Extern 1 (RegisterSpec UnionFind_ml.union) => Provide union_spec.

Open Scope nat_scope.

(* -------------------------------------------------------------------------- *)
(* Statements like in the paper. *)

Theorem make_spec' : forall N D R, card D < N ->
  App UnionFind_ml.make tt;
    (UF N D R \* \$ 1)
    (fun x => UF N (D \u \{x}) R \* \[x \notin D] \* \[R x = x]).
Proof using. intros. xapp~. hsimpl~. Qed.

Theorem find_spec' : forall N D R x, x \in D -> 
  App UnionFind_ml.find x;
    (UF N D R \* \$ (alpha N + 2))
    (fun y => UF N D R \* \[R x = y]).
Proof using. intros. xapp~. hsimpl~. Qed.

Theorem union_spec' : forall N D R x y, x \in D -> y \in D ->
  App UnionFind_ml.union x y; 
    (UF N D R \* \$(3*(alpha N)+6)) 
    (fun z => UF N D (fun w => If R w = R x \/ R w = R y then z else R w)
            \* \[z = R x \/ z = R y]).
Proof using. intros. xapp~. unfold link_by_rank_R. hsimpl~. Qed.

Theorem equiv_spec' : forall N D R x y, x \in D -> y \in D ->
  App UnionFind_ml.equiv x y;
    (UF N D R \* \$(2*(alpha N) + 5)) 
    (fun b => UF N D R \* \[b = true <-> R x = R y]).
Proof using. intros. xapp~. hsimpl~. subst r. rew_isTrue; autos*. Qed.



(* -------------------------------------------------------------------------- *)

