(* This module offers a union-find data structure based on disjoint set
   forests, with path compression and linking by rank. *)

(* -------------------------------------------------------------------------- *)

(* The rank of a vertex is defined as the maximum length, in edges, of an
   uncompressed path that leads to this vertex. In other words, the rank of
   [x] is the height of the tree rooted at [x] that would exist if we did not
   perform path compression. *)

type rank =
  int

(* The type [elem] represents a vertex in the union-find data structure.
   It is a primitive mutable reference. *)

type elem =
  content ref

(* The content of a vertex is either a pointer to a parent vertex (if the
   vertex has a parent) or a pair of a rank. *)

and content =
| Link of elem
| Root of rank 


(* -------------------------------------------------------------------------- *)

(* [make v] creates a new root of rank zero. *)

let make () : elem =
  ref (Root 0)

(* -------------------------------------------------------------------------- *)

(* [find x] finds the representative vertex of the equivalence class of [x].
   It does by following the path from [x] to the root. Path compression is
   performed (on the way back) by making every vertex along the path a
   direct child of the representative vertex. No rank is altered. *)

(* The function [find] is currently not exposed to the user, because [get]
   should be what the user needs in most situations. *)

let rec find x =
  match !x with
  | Root _ ->
      x
  | Link y ->
      let z = find y in
      x := Link z;
      z

(* -------------------------------------------------------------------------- *)

(* [equiv x y] determines whether the vertices [x] and [y] belong in the same
   equivalence class. It does so via two calls to [find] and a physical
   equality test. *)

let equiv x y =
  let x = find x in
  let y = find y in
  x == y


(* -------------------------------------------------------------------------- *)

(* [link x y] requires the vertices [x] and [y] to be roots. If they are
   the same vertex, it does nothing. Otherwise, it merges their equivalence
   classes by installing a link from one vertex to the other. *)

(* Linking is by rank: the smaller-ranked vertex is made to point to the
   larger. If the two vertices have the same rank, then an arbitrary choice
   is made, and the rank of the new root is incremented by one. *) 

let link x y =
  if x == y then x else
  match !x, !y with
  | Root rx, Root ry ->
      if rx < ry then begin
        x := Link y; y
      end else if rx > ry then begin
        y := Link x; x
      end else begin
        y := Link x; x := Root (rx+1); x
      end
  | _, _ ->
      assert false
  
(* -------------------------------------------------------------------------- *)

(* [union x y] is identical to [link s x y], except it does not require
   [x] and [y] to be roots. *)

let union x y =
  let x = find x in
  let y = find y in
  link x y

