Set Implicit Arguments.

Require Import LibTactics.
Require Import LibLogic.
Require Import LibEpsilon.
Require Import Omega.
Require Import LibFunc.
Generalizable Variables A.



(* -------------------------------------------------------------------------- *)

(* LibTactics local extension. *)

Hint Extern 1 (_ = _) => congruence : congruence.

(* FRANCOIS: this should only be used for terminal goals *)
Ltac unpack := jauto_set_hyps; intros.

Tactic Notation "branches" :=
  match goal with h: _ \/ _ |- _ => branches h end.

(* FRANCOIS investigate whether "cases_if" could work *)

Ltac by_cases_If :=
  match goal with |- context[If ?P then _ else _] =>
    tests : P
  end.

Ltac three_ways r1 r2 :=
  tests : (r1 = r2); [ | tests : (r1 < r2) ];
    (* TEMPORARY not reliable because tests: can produce only one sub-goal! *)
  repeat rewrite If_l in * by omega;
  repeat rewrite If_r in * by omega.

(* FRANCOIS : to reliably have 2 subgoals in "tests" you can do:
   destruct (classicT (r1 = r2)).  
*)

(* FRANCOIS: try replacing calls to "by_cases_on_fupdate" 
   (or the body of the definition below) 
   with "rewrite fupdate_def; case_if" *)

Ltac by_cases_on_fupdate :=
  match goal with |- context[fupdate ?f ?a ?b ?x] =>
    tests : (x = a);
    repeat rewrite fupdate_neq by eauto;
    repeat rewrite fupdate_eq by eauto
  end.


(* FRANCOIS: try using inverts where this is used *)
(* The tactic [injections] applies [injection] to the equality hypotheses that
   support it. *)

Ltac injections :=
  repeat match goal with
  | h: exist _ ?x _ = exist _ ?y _ |- _ =>
      assert (x = y); [ congruence | clear h ]
  | h: _ = _ |- _ =>
      first [
        (* Try to use [injection] on this hypothesis. *)
        injection h; clear h
        (* If this fails, set it aside. *)
      | revert h
      ]
  end;
  (* When done, re-introduce the hypotheses that were set aside. *)
  intros;
  (* Try substituting what can be substituted. *)
  try subst.



(* ------------------------------------------------------------------------- *)

(* LibRelation and LibPer *)

Require Import LibRelation.
Require Import LibPer.

(* The tactic [exploit_functional] looks for two hypotheses of
   the form [F x y] and [F x z], where [F] is functional, and
   uses them to prove the equality [y = z] and simplify the goal. *)
(* FRANCOIS: can you change your proofs to [functional_exploit F],
   which I will later integrate into TLC? *)

Create HintDb functional.

Ltac exploit_functional :=
  match goal with

  | h1: functional ?F, h2: ?F ?x ?y, h3: ?F ?x ?z |- _ =>
      assert (y = z); [
        solve [ eapply h1; eauto ]
      | clear h3; try subst y
      ]

  |                    h2: ?F ?x ?y, h3: ?F ?x ?z |- _ =>
      let f := fresh in
      assert (f: functional F); [
        solve [ eauto with functional ]
      | exploit_functional; clear f
      ]

  end.

(* Some lemmas that prove membership in a relation. *)

Hint Resolve prove_rel_union_left prove_rel_union_right prove_per_single : relations.


(* A derived lemma involving a step from rtclosure to tclosure *)

Lemma tclosure_intro_right:
  forall (A : Type) (R : binary A) y x z,
  rtclosure R x y -> R y z -> tclosure R x z.
Proof using.
  induction 1; eauto using tclosure_intro, tclosure_rtclosure with rtclosure.
Qed.

(* An explicit formulation of [rtclosure_trans], which can be used by [eauto]. *)

Lemma rtclosure_trans_explicit:
  forall (A : Type) (R : binary A) y x z,
  rtclosure R x y -> rtclosure R y z -> rtclosure R x z.
Proof using. eapply rtclosure_trans. Qed.


(* ------------------------------------------------------------------------- *)

(* Path *)

(* TEMPORARY: the definition of kpath could be replaced with a predicate over a list *) 

Section Closures.

Variables (A : Type) (R : binary A).

Inductive kpath : nat -> binary A :=
| KPathRefl:
    forall x,
    kpath 0 x x
| KPathStep:
    forall y x z k,
    R x y -> kpath k y z -> kpath (S k) x z.

Hint Constructors kpath : kpath.

Lemma kpath_rtclosure:
  forall k x y,
  kpath k x y ->
  rtclosure R x y.
Proof using.
  induction 1; eauto with rtclosure.
Qed.

Lemma rtclosure_kpath:
  forall x y,
  rtclosure R x y ->
  exists k, kpath k x y.
Proof using.
  induction 1; unpack; eauto with kpath.
Qed.

Lemma tclosure_kpath:
  forall x y,
  tclosure R x y ->
  exists k, kpath k x y /\ k > 0.
Proof using.
  induction 1.
  forwards [ k ? ]: rtclosure_kpath. eauto.
  exists (S k). split. eauto with kpath.
  omega.
Qed.

End Closures.

Hint Resolve kpath_rtclosure : rtclosure.


Require Import LibList.


(* ------------------------------------------------------------------------- *)


(* Intervals of the natural integers. *)
(* LATER: should be expressed on type fset *)
(* LATER: move to TLC, using TLC's comparison operator *)

Open Scope set_scope.
Require Import List LibList LibNat LibWf LibFix LibSet.
Close Scope comp_scope.

Definition interval_as_set (m n : nat) : set nat :=
  \set{ i | m <= i < n }.

Definition Interval_as_list f (m n:nat) :=
  If m >= n then nil else m::(f (S m) n).

Definition interval_as_list := FixFun2 Interval_as_list.

Lemma fix_interval_as_list : forall m n, 
  interval_as_list m n = Interval_as_list interval_as_list m n.
Proof using.
  applys~ (FixFun2_fix (measure2 (fun m n => (n - m)%nat))).
  applys measure2_wf. 
  introv IH. unfolds. case_if~. fequals.
  apply IH. simpls. nat_math. 
Qed.

Lemma interval_as_list_length : forall m n, m <= n ->
  length (interval_as_list m n) = n - m.
Proof using.
  introv. induction_wf IH: (nat_upto_wf n) m. intros Le.
  rewrite fix_interval_as_list. unfold Interval_as_list.
  case_if; rew_list. 
    nat_math.
    rewrite IH. nat_math. hnf. nat_math. nat_math.
Qed.

(* auxiliary lemma to help below *)
Lemma card_prove_eq' : forall A (E:set A) L n, 
  list_covers E L ->
  length L = n ->
  (forall L', list_covers E L' -> (length L' >= n)%nat) -> 
  card E = n.
Proof using.
  intros. subst. applys* card_prove_eq.
Qed.

Lemma covers_interval_as_set:
  forall m n,
  list_covers (interval_as_set m n) (interval_as_list m n).
Proof using.
  unfold interval_as_set.
  introv Hx. rewrite in_set_st_eq in Hx. gen Hx.
  induction_wf IH: (nat_upto_wf n) m.
  introv L. rewrite fix_interval_as_list. unfold Interval_as_list.
  case_if.
    false. nat_math.
    tests: (x = m). 
      auto.
      constructors. applys~ IH. unfolds. nat_math. nat_math.
Qed.

Lemma finite_interval_as_set:
  forall m n,
  finite (interval_as_set m n).
Proof using.
  intros. applys finite_prove_covers (interval_as_list m n).
  applys covers_interval_as_set.
Qed.

Lemma card_interval_as_set:
  forall m n, 
  m <= n ->
  card (interval_as_set m n) = n - m.
Proof using. 
  intros. applys card_prove_eq'.
    applys covers_interval_as_set.
    applys~ interval_as_list_length.
  gen H. induction_wf IH: (nat_upto_wf n) m. introv H C. 
  tests: (m = n). nat_math.
  forwards: IH (m+1) (Filter (<> m) L').
    hnf; nat_math.
    nat_math.
    intros x Hx. unfolds interval_as_set. rewrite in_set_st_eq in Hx.
     forwards R: C x.
       rewrite in_set_st_eq. nat_math.
       applys~ Filter_Mem. nat_math.
  lets: Filter_negated_predicates_length (=m) L'.
  forwards: Filter_eq_Mem_length m L'.
    applys C. unfold interval_as_set. rewrite in_set_st_eq. nat_math.
  nat_math.
Qed.



(* -------------------------------------------------------------------------- *)

Lemma le_refl : refl Peano.le.
Proof using. intros_all; omega. Qed.
Hint Resolve le_refl.



(* -------------------------------------------------------------------------- *)


Require Import LibMin.

Require Import LibOrder.


(* Tactics for mmin *)

Ltac mmin_spec t :=
  let h1 := fresh "existence" in
  let h2 := fresh "minimality" in
  (forwards [ h1 h2 ]: @mmin_spec t); [
    try reflexivity
  |
  | eauto with admits_lower_bound
  | eauto with bounded_has_minimal
  | eauto 2 ].


(* For this development, we use a definition of [mmax] and [Mmax]
   provaby equivalent to the one in TLC's library. *) 

Definition mmax `{Inhab A} (le:binary A) (P:A->Prop) :=
  mmin (flip le) P.

Lemma mmax_equiv : @mmax = @LibMin.mmax.
Proof using.
  extens. intros. unfold mmax. rewrite~ <- mmax_flip.
Qed.

Definition MMax `{Inhab A} `{Le A} := mmax le.



(* ------------------------------------------------------------------------- *)
(* Extension to LibSet *)

Require Import LibSet.

Section Autorewrite.
Variables (A : Type).
Implicit Types x y : A.
Implicit Types E F : set A.

Lemma notin_eq' : forall x E, (x \notin E) = ~ (x \in E).
Proof using. intros. apply notin_eq. Qed.

End Autorewrite.

Hint Rewrite notin_eq' : set_norm.


(* ------------------------------------------------------------------------- *)


(* LibBag or LibSet *)

Require Import LibStruct.
Require Import LibSet.
Require Export Coq.Classes.Morphisms. (* for rewriting *)


(* In my opinion, \notin should be just a notation! *)

(* FRANCOIS : inline, except if used for autorewrite *)
Lemma notin_eq :
  forall A (E : set A) x,
  (x \notin E) = (~ x \in E).
Proof using. intros. rewrite notin_eq. auto. Qed. 

(* FRANCOIS : inline, except if used for autorewrite *)
Lemma disjoint_single_left_eq:
  forall A (E : set A) x,
  \{x} \# E = ~ x \in E.
Proof using. intros. apply disjoint_single_l_eq. Qed.

(* FRANCOIS : inline, except if used for autorewrite *)
Lemma disjoint_single_right_eq:
  forall A (E : set A) x,
  E \# \{x} = ~ x \in E.
Proof using. intros. apply disjoint_single_r_eq. Qed.

(* FRANCOIS : inline, except if used for autorewrite *)
Lemma subset_single_eq:
  forall A (E : set A) x,
  \{x} \c E = x \in E.
Proof using. intros. apply single_incl_r_eq. Qed.

(* FRANCOIS : inline, except if used for autorewrite *)
Lemma disjoint_implies_empty_inter:
  forall A (E F : set A),
  E \# F ->
  E \n F = \{}.
Proof using. intros. apply~ inter_disjoint. Qed.

(* Note: derivable from Incl_union_l *)
Lemma incl_insert:
  forall A (E : set A) x,
  E \c (E \u \{x}).
Proof using. set_prove. Qed.

(* FRANCOIS : inline, except if used for autorewrite *)
Lemma prove_not_in_empty:
  forall A (a : A),
  a \notin (\{} : set A).
Proof using. set_prove. Qed.

Lemma prove_notin_inter:
  forall A (E F : set A) x,
  x \notin E \/ x \notin F ->
  x \notin E \n F.
Proof using. set_prove. Qed.
  (* TODO: add support for this: intros. rewrite notin_inter_eq. *)

(* Note: derivable form in_union_r and in_single *)
Lemma in_insert:
  forall A (E : set A) x,
  x \in E \u \{x}.
Proof using. set_prove. Qed.


Lemma inter_union_disjoint_right:
  forall A (E F G : set A),
  F \# G ->
  (E \u F) \n G = (E \n G).
Proof using. set_prove. Qed.

Lemma inter_union_subset_right:
  forall A (E F G : set A),
  F \c G ->
  (E \u F) \n G = (E \n G) \u F.
Proof using. set_prove. Qed.

(* TEMPORARY should set up [Proper] instances for rewriting *)
Lemma inter_covariant:
  forall A (E E' F F' : set A),
  E \c E' ->
  F \c F' ->
  (E \n F) \c (E' \n F').
Proof using. set_prove. Qed.

(* Disjointness. *)

(* FRANCOIS : inline, except if used for automation *)
Lemma prove_disjoint : 
  forall (A : Type) (E F : set A),
  (forall x, x \in E -> x \in F -> False) ->
  E \# F.
Proof using. set_prove. Qed.
  (* TODO: add support for this: intros. applys Disjoint_prove. *)

(* FRANCOIS : inline, except if used for automation *)
Lemma remove_incl:
  forall A (E F : set A),
  (E \- F) \c E.
Proof using. intros. apply~ remove_incl. Qed.

(* FRANCOIS : inline, except if used for automation *)
Lemma remove_disjoint:
  forall A (E F : set A),
  F \# (E \- F).
Proof using. intros. apply~ remove_disjoint. Qed.

(* ------------------------------------------------------------------------- *)


Hint Rewrite disjoint_single_left_eq disjoint_single_right_eq subset_single_eq
@notin_eq @in_empty_eq @in_single_eq @in_union_eq @in_inter_eq @in_remove_eq : rew_set.

Hint Resolve finite_empty finite_single finite_union
  finite_inter finite_incl finite_remove : finite.

Hint Resolve finite_image : finite.

Tactic Notation "rew_set" "in" "*" :=
  autorewrite with rew_set in *;
  eauto with typeclass_instances.

(* ------------------------------------------------------------------------- *)

Program Instance Reflexive_incl A :
  Reflexive (@LibBag.incl _ (incl_inst A)).
Next Obligation.
  eapply incl_refl; eauto.
Qed.

Program Instance Transitive_incl A :
  Transitive (@LibBag.incl _ (incl_inst A)).
Next Obligation.
  eapply incl_trans; eauto.
Qed.

Program Instance Proper_inter_incl (A : Type) :
  Proper (@LibBag.incl _ (incl_inst A) ++>
          @LibBag.incl _ (incl_inst A) ++>
          @LibBag.incl _ (incl_inst A))
  (@LibBag.inter _ (inter_inst A)).
Next Obligation.
  repeat intro. eauto using inter_covariant.
Qed.

Goal
  forall A (E F G H : set A),
  E \c F ->
  F \n G \c H ->
  E \n G \c H.
Proof using.
  introv h1 h2.
  rewrite h1. rewrite h2. reflexivity.
Qed.

Tactic Notation "set_in" :=
  match goal with
  | h: _ \in \{} |- _    => set_in h
  | h: _ \in \{_} |- _   => set_in h
  | h: _ \in _ \u _ |- _ => set_in h
  end.


(* ------------------------------------------------------------------------- *)


Global Instance monoid_commutative_plus_zero:
  Monoid_commutative (monoid_ plus 0).
Proof using.
  constructor. typeclass. intros_all; omega.
Qed.




(* Folding over finite sets. *)
(* LATER: should be expressed on type fset *)
(* TODO: could move to TLC *)


Lemma fold_isolate :
  forall A (E : set A) x,
  finite E ->
  x \in E ->
  forall B (m : monoid_def B),
  Monoid_commutative m ->
  forall (f : A -> B),
  fold m f E = monoid_oper m (f x) (fold m f (E \- \{x})).
Proof using.
  intros.
  (* Separate [E] into the singleton [\{x}] union the rest. *)
  rewrite (set_isolate E x) at 1 by eauto.
  (* Note that [f x] is the result of folding [f] over the singleton [\{x}]. *)
  erewrite <- (fold_single f x) by typeclass.
  (* Conclude. *)
  eapply fold_union; eauto using remove_disjoint with finite.
Qed.

Require Import Coq.Classes.Morphisms.

Lemma list_fold_pointwise:
  forall B (m : monoid_def B) (leB : B -> B -> Prop),
  Monoid m ->
  refl leB ->
  Proper (leB ++> leB ++> leB) (monoid_oper m) ->
  forall A (L : list A),
  forall (f f' : A -> B),
  (forall x, Mem x L -> leB (f x) (f' x)) ->
  leB (LibList.fold m f L) (LibList.fold m f' L).
Proof using.
  introv HM HR HP. induction L; introv HL.
  do 2 rewrite LibList.fold_nil. applys HR.
  do 2 rewrite LibList.fold_cons. apply HP. applys~ HL. applys~ IHL.
Qed.

Lemma fold_pointwise:
  forall B (m : monoid_def B) (leB : B -> B -> Prop),
  Monoid m ->
  refl leB ->
  Proper (leB ++> leB ++> leB) (monoid_oper m) ->
  forall A (E : set A),
  finite E ->
  forall (f f' : A -> B),
  (forall x, x \in E -> leB (f x) (f' x)) ->
  leB (fold m f E) (fold m f' E).
Proof using.
  intros. do 2 rewrite fold_def.
  applys~ list_fold_pointwise.
  intros x. forwards~ (_&EQ): finite_list_repr E. rewrite (EQ x). auto.
Qed.


(* ------------------------------------------------------------------------- *)

(* [confined D F] could be alternatively defined as:
   [confined D F := dom F = D /\ img F = D].
   where TLC would include 
   [dom F = { x | exists y, F x y }]
   [img F = { y | exists x, F x y }].
  *)

(* A relation [F] is confined in a domain [D] iff [F x y]
   implies [x \in D] and [y \in D]. In other words, [F] is
   a subset of [D^2]. *)

Definition confined (A : Type) (D : set A) (F : binary A) :=
  forall x y, F x y -> x \in D /\ y \in D.

Lemma use_confined_left:
  forall (A : Type) (D : set A) F,
  confined D F ->
  forall x y,
  F x y ->
  x \in D.
Proof using.
  unfold confined. introv h. intros.
  forwards: h. eauto. tauto.
Qed.

Lemma use_confined_right:
  forall (A : Type) (D : set A) F,
  confined D F ->
  forall x y,
  F x y ->
  y \in D.
Proof using.
  unfold confined. introv h. intros.
  forwards: h. eauto. tauto.
Qed.

Lemma confined_union:
  forall (A : Type) D (F1 F2 : binary A),
  confined D F1 ->
  confined D F2 ->
  confined D (LibRelation.union F1 F2).
Proof using.
  unfold confined, LibRelation.union.
  intros. branches; eauto.
Qed.

Lemma confined_per_single:
  forall (A : Type) D (x y : A),
  x \in D ->
  y \in D ->
  confined D (per_single x y).
Proof using.
  unfold confined, per_single. intros; unpack; subst. eauto.
Qed.

Lemma confined_stclosure:
  forall (A : Type) D (F : binary A),
  confined D F ->
  confined D (stclosure F).
Proof using.
  unfold confined. introv h. induction 1; unpack; eauto.
Qed.

Hint Resolve use_confined_left use_confined_right confined_union
confined_per_single confined_stclosure : confined.

Definition confine (A : Type) (D : set A) (F : binary A) : binary A :=
  fun x y => x \in D /\ y \in D /\ F x y.

Ltac binary_extensional :=
  intros; unfold LibRelation.empty;
  eapply binary_extensional;
  split; intros; unpack; rew_set in *; (* dependency on LibSet, not great *)
    try solve [ tauto | false ].

(* Confining a relation [F] to the empty set yields the empty relation. *)

Lemma confine_empty:
  forall A (F : binary A),
  confine \{} F = (@LibRelation.empty _).
Proof using.
  unfold confine. binary_extensional.
Qed.

Lemma confine_equiv_per:
  forall (A : Type) (F : binary A) (D : set A),
  equiv F ->
  per (confine D F).
Proof using.
  unfold confine. inversion 1.
  constructor; repeat intro; intuition eauto.
Qed.

Lemma confine_union_left:
  forall (A : Type) (D : set A) R S,
  confined D S ->
  LibRelation.union (confine D R) S =
  confine D (LibRelation.union R S).
Proof using.
  (* Slightly ad hoc statement and proof. *)
  unfold confined, confine, LibRelation.union.
  introv hc. eapply binary_extensional; split; intros. 
  { branches; unpack.
      eauto.
      forwards: hc. eauto. unpack. eauto. }
  { unpack. branches; eauto. }
Qed.

(* By construction, [confine D F] is confined within [D]. *)

Lemma confined_confine:
  forall (A : Type) (F : binary A) (D : set A),
  confined D (confine D F).
Proof using.
  unfold confined, confine. tauto.
Qed.

Lemma incl_confine:
  forall (A : Type) (F : binary A) (D : set A),
  LibRelation.incl (confine D F) F.
Proof using.
  unfold LibRelation.incl, confine. intuition eauto.
Qed.

Hint Resolve confined_confine incl_confine : confined.


(* ------------------------------------------------------------------------- *)

(** *

   [sticky D R] could be renamed as [is_component D R], 
   in the graph terminology

*)

Definition sticky (A : Type) (D : set A) (R : binary A) :=
  forall x y, R x y -> (x \in D <-> y \in D).

Lemma use_sticky_left:
  forall (A : Type) (D : set A) (R : binary A) x y,
  x \in D ->
  R x y ->
  sticky D R ->
  y \in D.
Proof using.
  unfold sticky. introv ? ? h.
  forwards: h. eauto. tauto.
Qed.

Lemma use_sticky_right:
  forall (A : Type) (D : set A) (R : binary A) x y,
  y \in D ->
  R x y ->
  sticky D R ->
  x \in D.
Proof using.
  unfold sticky. introv ? ? h.
  forwards: h. eauto. tauto.
Qed.

Hint Resolve use_sticky_left use_sticky_right : sticky.

Lemma sticky_per_single:
  forall (A : Type) (D : set A) x y,
  (x \in D <-> y \in D) ->
  sticky D (per_single x y).
Proof using.
  unfold sticky, per_single.
  intros. unpack. subst. eauto.
Qed.

Lemma sticky_union:
  forall (A : Type) (D : set A) (R S : binary A),
  sticky D R ->
  sticky D S ->
  sticky D (LibRelation.union R S).
Proof using.
  unfold sticky, LibRelation.union.
  intros. branches; eauto.
Qed.

Lemma sticky_stclosure:
  forall (A : Type) (D : set A) (R : binary A),
  sticky D R ->
  sticky D (stclosure R).
Proof using.
  unfold sticky. induction 2. eauto. tauto. tauto.
Qed.

Hint Resolve sticky_per_single sticky_union sticky_stclosure : sticky.

Lemma confine_stclosure:
  forall (A : Type) (D : set A) (R : binary A),
  sticky D R ->
  confine D (stclosure R) = stclosure (confine D R).
Proof using.
  intros.
  assert (forall x y, stclosure R x y -> x \in D -> y \in D).
    { intros. forwards h: sticky_stclosure x y; eauto. tauto. }
  eapply binary_extensional. split.
  (* One direction. *)
  { unfold confine. introv (? & ? & h). gen h. induction 1.
      { eapply stclosure_step. eauto. }
      { eapply stclosure_sym; eauto. }
      { eapply stclosure_trans; eauto. }
  }
  (* The other direction. *)
  { intros. repeat split; eauto with confined.
    eapply stclosure_le; [ | eauto ].
    eauto with confined. }
Qed.





(* ------------------------------------------------------------------------- *)

(* NOT USED *)


(* TEMPORARY clumsy
  Definition inclusion_of_finite_sets A (E F : set A) :=
    E \c F /\ finite F.

  Program Instance Proper_card_incl (A : Type) :
    Proper (@inclusion_of_finite_sets A ++> le) card.
  Next Obligation.
    (* @LibBag.incl _ (incl_inst A) *)
    unfold inclusion_of_finite_sets. repeat intro. intuition eauto using card_incl_le.
  Qed.
*)



